unit SpiFilter;

interface

uses Windows, spi_api, SUtil, ImgFilter;

type
  TArchiveInfo = record
    FileName: string;
    FileSize: Cardinal;
    CompSize: Cardinal;
    DateTime: TFileTime;
    RawData: Pointer;
  end;

  TArchive = class(TObject)
  private
    FOpened: Boolean;
    FFileName: string;
    Handle: THandle;
    HInfo: HLOCAL;
    SpiFileInfoTop: PFileInfo;
    SpiFileInfo: PFileInfo;
    GetArchiveInfo: TGetArchiveInfo;
    GetFile: TGetFile;
    function GetInfo(const Name: string): TArchiveInfo;
  public
    constructor Create(const PluginFileName: string; const AFileName: string);
    destructor Destroy; override;
    function EachName(var Name: string): Boolean;
    function Extract(const Name: string; Buf: PChar; BufSize: Longint): Boolean;
    property FileName: string read FFileName;
    property Info[const Name: string]: TArchiveInfo read GetInfo; default;
    property Opened: Boolean read FOpened;
  end;

  TSpiFilter = class(TImgFilter)
  private
    FFileName: string;
    GetPluginInfo: TGetPluginInfo;
    IsSupported: TIsSupported;
    GetPictureInfo: TGetPictureInfo;
    GetPicture: TGetPicture;
    ConfigurationDlg: TConfigurationDlg;
    function GetErrMsg(Code: Integer): string;
  public
    constructor Create(const PluginFileName: string);
    destructor Destroy; override;
    function Test(Buf: PChar; BufSize: Integer): Boolean; override;
    function GetInfo(Buf: PChar; BufSize: Longint; var ImgInfo: TImgInfo): Boolean; override;
    function Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar;
      OutBufSize: Longint; BitmapInfo: PBitmapInfo): Boolean; override;
    function IsArchive: Boolean;
    function ShowConfigDlg(hWnd: HWND): Boolean;
    property FileName: string read FFileName;
  end;


implementation

const
  SpiErrMsg: array[0..10] of string = (
    '',
    'user cancel',
    'unknown format',
    'broken data',
    'allocate memory error',
    'memory error',
    'file read error',
    'cannot create window',
    'internal error',
    'file write error',
    'end of file error'
  );

constructor TSpiFilter.Create(const PluginFileName: string);
var
  buf: array[0..255] of Char;
begin
  Handle := LoadLibrary(PChar(PluginFileName));
  if Handle <> 0 then
  begin
    FFileName := PluginFileName;
    GetPluginInfo := GetProcAddress(Handle, 'GetPluginInfo');
    IsSupported := GetProcAddress(Handle, 'IsSupported');
    GetPictureInfo := GetProcAddress(Handle, 'GetPictureInfo');
    GetPicture := GetProcAddress(Handle, 'GetPicture');
    ConfigurationDlg := GetProcAddress(Handle, 'ConfigurationDlg');
    if Assigned(GetPluginInfo) then
    begin
      GetPluginInfo(1, buf, SizeOf(buf));
      FName := buf;
    end;
  end;
end;

destructor TSpiFilter.Destroy;
begin
  if Handle <> 0 then
    FreeLibrary(Handle);
end;

function TSpiFilter.GetErrMsg(Code: Integer): string;
begin
  Result := IStr(Code);
  if (Code >= 0) and (Code <= High(SpiErrMsg)) then
    Result := Result + ':' + SpiErrMsg[code];
end;

function TSpiFilter.Test(Buf: PChar; BufSize: Integer): Boolean;
var
  tmpbuf: array[0..2047] of Char;
begin
  // vOCobt@j󂷂邩Ȃ̂ŃRs[Ăn
  if BufSize > SizeOf(tmpbuf) then
    BufSize := SizeOf(tmpbuf);
  FillChar(tmpbuf, SizeOf(tmpbuf), 0);
  Move(Buf^, tmpbuf, BufSize);
  Result := Assigned(IsSupported) and (IsSupported('', DWORD(@tmpbuf)) <> 0);
end;

function TSpiFilter.GetInfo(Buf: PChar; BufSize: Longint; var ImgInfo: TImgInfo): Boolean;
var
  picinfo: TPictureInfo;
  r: Integer;
begin
  if not Assigned(GetPictureInfo) then
  begin
    FErrorMsg := 'Get address of GetPictureInfo failed.';
    Result := False;
    Exit;
  end;

  // PicInfoSZbgȂvOC΍̂ߕK0NA
  FillChar(picinfo, SizeOf(picinfo), 0);
  r := GetPictureInfo(Buf, BufSize, 1, picinfo);
  if r <> 0 then
  begin
    FErrorMsg := 'GetPictureInfo failed. (spi errcode='+GetErrMsg(r)+')';
    Result := False;
    Exit;
  end;

  with ImgInfo do
  begin
    Width := picinfo.width;
    Height := picinfo.height;
    Bpp := picinfo.colorDepth;
    MimeType := '';
    FormatName := '';
    if picinfo.hInfo = 0 then
      Comment := ''
    else begin
      SetString(Comment, PChar(LocalLock(picinfo.hInfo)),
        LocalSize(picinfo.hInfo)-1); // -1: null
      LocalUnLock(picinfo.hInfo);
      LocalFree(picinfo.hInfo);
    end;
  end;

  Result := True;
end;

function ProgressCallback(nNum, nDemon: Integer; lData : Longint): Integer; stdcall;
begin
  Result := 0;
end;

function TSpiFilter.Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar;
  OutBufSize: Longint; BitmapInfo: PBitmapInfo): Boolean;
const
  MaxBmiSize = SizeOf(TBitmapInfoHeader)+SizeOf(TRGBQuad)*256;
var
  hbinfo, hbm: HLOCAL;
  r, size: Integer;
begin
  if not Assigned(GetPicture) then
  begin
    FErrorMsg := 'Get address of GetPicture failed.';
    Result := False;
    Exit;
  end;

  // ProgressCallbacknilƗvOC΍̂߃_~[֐n
  r := GetPicture(InBuf, InBufSize, 1, hbinfo, hbm, ProgressCallback, 0);
  if r <> 0 then
  begin
    FErrorMsg := 'GetPicture failed. (spi errcode='+GetErrMsg(r)+')';
    Result := False;
    Exit;
  end;

  size := LocalSize(hbinfo);
  if size > MaxBmiSize then
    size := MaxBmiSize;
  Move((LocalLock(hbinfo))^, BitmapInfo^, size);
  LocalUnLock(hbinfo);
  LocalFree(hbinfo);

  size := LocalSize(hbm);
  if size > OutBufSize then
    size := OutBufSize;
  Move((LocalLock(hbm))^, OutBuf^, size);
  LocalUnLock(hbm);
  LocalFree(hbm);

  Result := True;
end;

function TSpiFilter.IsArchive: Boolean;
var
  ver: array[0..15] of Char;
begin
  FillChar(ver, SizeOf(ver), 0);
  if Assigned(GetPluginInfo) then
    GetPluginInfo(0, ver, SizeOf(ver));
  Result := Copy(ver, 1, 4) = '00AM';
end;

function TSpiFilter.ShowConfigDlg(hWnd: HWND): Boolean;
var
  r: Integer;
begin
  if Assigned(ConfigurationDlg) then
  begin
    r := ConfigurationDlg(hWnd, 1);
    if r = 0 then
    begin
      Result := True;
      Exit;
    end;
    if r > 0 then
    begin
      FErrorMsg := 'ConfigurationDlg failed. (spi errcode='+GetErrMsg(r)+')';
      Result := False;
      Exit;
    end;
  end;
  FErrorMsg := 'Configuration dialog is not implemented.';
  Result := False;
end;


{ TArchive }

constructor TArchive.Create(const PluginFileName: string; const AFileName: string);
begin
  FFileName := AFileName;
  Handle := LoadLibrary(PChar(PluginFileName));
  if Handle <> 0 then
  begin
    GetArchiveInfo := GetProcAddress(Handle, 'GetArchiveInfo');
    GetFile := GetProcAddress(Handle, 'GetFile');
    if Assigned(GetArchiveInfo) then
    begin
      if GetArchiveInfo(PChar(FFileName), 0, 0, HInfo) = 0 then
      begin
        SpiFileInfoTop := LocalLock(HInfo);
        SpiFileInfo := SpiFileInfoTop;
        FOpened := True;
      end;
    end;
  end;
end;

destructor TArchive.Destroy;
begin
  if HInfo <> 0 then
  begin
    LocalUnlock(HInfo);
    LocalFree(HInfo);
  end;
  if Handle <> 0 then
    FreeLibrary(Handle);
end;

function TArchive.GetInfo(const Name: string): TArchiveInfo;
  function time_tToFileTime(t: Longint): TFileTime;
  var
    st: TSystemTime;
    ft: TFileTime;
  begin
    st.wYear := 1970;
    st.wMonth := 1;
    st.wDay := 1;
    st.wHour := 0;
    st.wMinute := 0;
    st.wSecond := 0;
    st.wMilliseconds := 0;
    SystemTimeToFileTime(st, ft);
    Comp(Result) := Comp(ft) + t * 10000000.0;
  end;
var
  spiinfo: PFileInfo;
begin
  spiinfo := SpiFileInfoTop;
  while spiinfo.method[0] <> #0 do
    if Name = spiinfo.path + spiinfo.filename then
    begin
      Result.FileName := Name;
      Result.FileSize := spiinfo.filesize;
      Result.CompSize := spiinfo.compsize;
      Result.DateTime := time_tToFileTime(spiinfo.timestamp);
      Result.RawData := spiinfo;
      Exit;
    end else
      Inc(spiinfo);
  Result.FileName := '';
  Result.RawData := nil;
end;

function TArchive.EachName(var Name: string): Boolean;
begin
  Result := SpiFileInfo.method[0] <> #0;
  if Result then
    Name := SpiFileInfo.path + SpiFileInfo.filename;
  Inc(SpiFileInfo);
end;

function TArchive.Extract(const Name: string; Buf: PChar; BufSize: Longint): Boolean;
var
  info: PFileInfo;
  hdst: HLOCAL;
  p: PChar;
  fsize: Integer;
begin
  Result := False;
  info := GetInfo(Name).RawData;
  if info = nil then
    Exit;

  // TODO: check error
  GetFile(PChar(FFileName), info.position, PChar(@hdst), $100, nil, 0);
  // TODO: check error
  p := LocalLock(hdst);
  fsize := LocalSize(hdst);
  if fsize > BufSize then
    fsize := BufSize;
  Move(p^, Buf^, fsize);
  LocalUnLock(hdst);
  LocalFree(hdst);
end;

end.
